import argparse
import json
import logging
import sys
from datetime import datetime, timedelta
from pathlib import Path
from time import sleep
from typing import Callable, Dict, List, Optional, Tuple, Union

import paho.mqtt.client as mqtt
from yaml import FullLoader, dump, load

from miqro import ha_sensors


class Loop:
    fn: Callable
    interval: timedelta
    next_call: Optional[datetime] = None
    stat_call_count: int = 0
    stat_cumulative_duration: float = 0.0

    def __init__(self, fn: Callable, interval: timedelta, start: bool = True):
        self.fn = fn

        if not isinstance(interval, timedelta):
            raise Exception("interval must be provided as timedelta!")

        self.interval = interval
        if start:
            self.next_call = datetime.now()

    def run_if_needed(self, instance: "Service") -> Optional[datetime]:
        now = datetime.now()
        if self.next_call and now >= self.next_call:
            if self.fn(instance) is not False:
                self.next_call = now + self.interval
                self.stat_call_count += 1
                self.stat_cumulative_duration += (datetime.now() - now).total_seconds()
                return self.next_call
            else:
                self.stop()
        return None

    def start(self, delayed=False):
        if delayed:
            self.next_call = datetime.now() + self.interval
        else:
            self.next_call = datetime.now()

    def stop(self):
        self.next_call = None

    def restart(self, delayed=False):
        self.start(delayed=delayed)

    def get_remaining(self) -> Optional[timedelta]:
        if self.next_call is None:
            return None
        return self.next_call - datetime.now()

    def stat_reset(self):
        self.stat_call_count = 0
        self.stat_cumulative_duration = 0.0

    def stat_get(self) -> Tuple[int, float, float, bool]:
        average_call_duration = (
            0.0
            if not self.stat_call_count
            else (self.stat_cumulative_duration / self.stat_call_count)
        )
        try:
            load = average_call_duration / self.interval.total_seconds()
        except ZeroDivisionError:
            load = 0.0
        is_critical = load > 1.0
        return self.stat_call_count, average_call_duration, load, is_critical

    def __str__(self):
        return f"Loop({self.fn.__name__})"


# args/kwargs can be anything that the constructor of timedelta accepts
def loop(*args, **kwargs):
    # See https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class
    class class_decorator:
        def __init__(self, fn):
            self.fn = fn

        def __set_name__(self, owner: "Service", name):
            if not owner.PREPARED_LOOPS:
                owner.PREPARED_LOOPS = []
            owner.PREPARED_LOOPS.append((self.fn, timedelta(*args, **kwargs)))

    return class_decorator


def handle(topic_ext):
    # See https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class
    class class_decorator:
        def __init__(self, fn):
            self.fn = fn

        def __set_name__(self, owner: "Service", name):
            if not owner.CLASS_MQTT_HANDLERS:
                owner.CLASS_MQTT_HANDLERS = []
            owner.CLASS_MQTT_HANDLERS.append((topic_ext, self.fn))

    return class_decorator


def handle_global(topic):
    # See https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class
    class class_decorator:
        def __init__(self, fn):
            self.fn = fn

        def __set_name__(self, owner: "Service", name):
            if not owner.CLASS_MQTT_GLOBAL_HANDLERS:
                owner.CLASS_MQTT_GLOBAL_HANDLERS = []
            owner.CLASS_MQTT_GLOBAL_HANDLERS.append((topic, self.fn))

    return class_decorator


def accept_json(fn):
    def actual_fn(self, arg):
        return fn(self, **json.loads(arg))

    return actual_fn


class State:
    """Store data in a YAML file."""

    service: "Service"

    DATA_ROOT = Path("/var/lib/miqro/data")

    def __init__(self, service) -> None:
        self.service = service
        self._file = self.DATA_ROOT / (service.SERVICE_NAME + ".yaml")

        try:
            if not self._file.exists():
                self._file.parent.mkdir(parents=True, exist_ok=True)
                self._data = {}
            else:
                with self._file.open() as f:
                    self._data = load(f, Loader=FullLoader)
        except PermissionError as e:
            service.log.error(e)
            self._data = {}

        self.service.log.debug(f"State: Loaded {self._data}")

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def set_path(self, *keys, value):
        self.service.log.debug(f"State: Setting {keys} to {value}")
        d = self._data
        for key in keys[:-1]:  # -1 because we don't want to set the last key
            if not key in d:
                d[key] = {}
            d = d[key]

        d[keys[-1]] = value

    def get_path(self, *keys, default):
        d = self._data
        for key in keys:
            if not key in d:
                self.service.log.debug(f"State: {key} not found, returning '{default}'")
                return default
            d = d[key]

        self.service.log.debug(f"State: {keys} found, returning {d}")
        return d

    def save(self):
        self.service.log.debug(f"State: Saving {self._data}")
        with self._file.open("w") as f:
            dump(self._data, f)


class Service:
    SERVICE_NAME: str = "none"
    CONFIG_FILE_PATHS: List[Path] = [Path("miqro.yml"), Path("/etc/miqro.yml")]
    JSON_FLOAT_PRECISION: int = 4
    MAX_LOOP_INTERVAL: float = 0.2
    PREPARED_LOOPS: List[Tuple[Callable, timedelta]] = []
    LOOPS: Optional[List[Loop]] = None
    CLASS_MQTT_HANDLERS: List[Tuple[str, Callable]] = []
    CLASS_MQTT_GLOBAL_HANDLERS: List[Tuple[str, Callable]] = []
    MQTT_ONLINE_UPDATE_INTERVAL: int = 180

    PAYLOAD_ON = 1
    PAYLOAD_OFF = 0

    ha_devices: List[ha_sensors.Device] = []
    ha_entities: List[ha_sensors.Entity] = []  # only entities without device

    USE_STATE_FILE = False

    QOS_MAX_ONCE = 0
    QOS_AT_LEAST_ONCE = 1
    QOS_EXACTLY_ONCE = 2

    log: logging.Logger
    config: Dict
    service_config: Dict
    data_topic_prefix: str
    mqtt_client: mqtt.Client
    stop = False
    is_connected = False
    mqtt_handlers: List[Tuple[str, Callable]]
    mqtt_global_handlers: List[Tuple[str, Callable]]
    state: Optional[State] = None

    def __init__(
        self,
        add_config_file_path=None,
        log_level=logging.DEBUG,
        mqtt_client_cls=mqtt.Client,
        state_cls=State,
    ):
        self._prepare_logger(log_level)
        self._read_config(add_config_file_path)

        self.last_key_values = {}

        self.mqtt_client = mqtt_client_cls(self.SERVICE_NAME)
        if "auth" in self.config:
            self.mqtt_client.username_pw_set(**self.config["auth"])
        if "tls" in self.config:
            self.mqtt_client.tls_set(**self._make_tls_config(self.config["tls"]))
        self.mqtt_client.on_connect = self._on_connect
        self.mqtt_client.on_disconnect = self._on_disconnect
        self.mqtt_client.on_message = self._on_message
        self.mqtt_client.enable_logger(self.mqtt_log)
        self.mqtt_client.will_set(self.willtopic, "0", retain=True)
        self.mqtt_client.connect_async(**self.config["broker"])

        self.enabled = True

        self.mqtt_handlers = [h for h in self.CLASS_MQTT_HANDLERS]
        self.mqtt_handlers.append(("enabled", self._on_enable))
        self.mqtt_global_handlers = [h for h in self.CLASS_MQTT_GLOBAL_HANDLERS]

        if self.USE_STATE_FILE:
            self.state = state_cls(self)

        self._create_loops()

        self.log.info("started")

    def __str__(self):
        return self.SERVICE_NAME

    def _make_tls_config(self, config):
        # cert_reqs and tls_version are strings pointing to properties in
        # the ssl module - parse from string to property!
        import ssl

        return {
            "ca_certs": config.get("ca_certs", None),
            "certfile": config.get("certfile", None),
            "keyfile": config.get("keyfile", None),
            "cert_reqs": getattr(ssl, config.get("cert_reqs", "CERT_REQUIRED")),
            "tls_version": getattr(ssl, config.get("tls_version", "PROTOCOL_TLS")),
            "ciphers": config.get("ciphers", None),
        }

    def add_loop(self, loop):
        if not self.LOOPS:
            self.LOOPS = []

        self.LOOPS.append(loop)
        return loop

    def _create_loops(self):
        self.add_loop(
            Loop(
                self._update_online_status,
                timedelta(seconds=self.MQTT_ONLINE_UPDATE_INTERVAL),
            )
        )
        for fn, interval in self.PREPARED_LOOPS:
            self.add_loop(Loop(fn, interval))

    def _prepare_logger(self, log_level):
        if not logging.getLogger().hasHandlers():
            log_handler = logging.StreamHandler(sys.stderr)
            log_handler.setFormatter(
                logging.Formatter("%(asctime)s  %(name)s  %(levelname)s \t%(message)s")
            )
            log_handler.setLevel(logging.DEBUG)
            logging.getLogger().addHandler(log_handler)
        logging.getLogger().setLevel(log_level)

        self.log = logging.getLogger(self.SERVICE_NAME + ".main")
        self.mqtt_log = logging.getLogger(self.SERVICE_NAME + ".mqtt")
        self.mqtt_log.setLevel(logging.INFO)

    def _read_config(self, add_config_file_path=None):
        paths = self.CONFIG_FILE_PATHS
        if add_config_file_path:
            paths.insert(0, Path(add_config_file_path))

        for path in paths:
            if path.exists():
                self.log.debug(f"Using configuration file at {path}")
                with path.open("r") as f:
                    self.config = load(f, Loader=FullLoader)
                    break
            else:
                self.log.debug(f"NOT using configuration file at {path}")
        else:
            raise Exception(
                "No MIQRO config file found; searched paths: "
                + ", ".join(map(str, self.CONFIG_FILE_PATHS))
            )

        if self.SERVICE_NAME not in self.config.get("services", {}):
            self.log.warning(
                f"Service configuration for {self.SERVICE_NAME} not found in 'services' section of configuration file {path}. Using empty configuration."
            )
            self.service_config = {}
        else:
            self.service_config = self.config["services"][self.SERVICE_NAME]

        self.data_topic_prefix = self.service_config.get(
            "data_topic", f"service/{self.SERVICE_NAME}/"
        )
        self.willtopic = self.data_topic_prefix + "online"

    def _on_connect(self, client, userdata, flags, rc):
        if rc != 0:
            self.log.error(f"MQTT connection failed with code {rc}")
            return
        self.log.info(f"MQTT connected, client={client}, userdata={userdata}, rc={rc}")
        self.is_connected = True
        self.log.info(f"Subscribing to ...")

        for topic, _ in self._all_handlers():
            self.log.info(f"  - {topic}")
            client.subscribe(topic)

        self.mqtt_client.publish(self.willtopic, "1", retain=True)

        self._publish_ha_discovery()

    def _publish_ha_discovery(self):
        if "ha_discovery_prefix" in self.config:
            prefix = self.config["ha_discovery_prefix"]
        else:
            prefix = "homeassistant"

        for device in self.ha_devices:
            device.publish_discovery(prefix)
        for entity in self.ha_entities:
            entity.publish_discovery(prefix)

    def _on_disconnect(self, client, userdata, rc):
        self.log.warning(f"MQTT disconnected, rc={rc}")
        self.is_connected = False

    def add_handler(self, topic, handler):
        self.mqtt_handlers.append((topic, handler))

        if self.is_connected:
            self.log.info(f"Subscribing to {self.data_topic_prefix + topic}")
            self.mqtt_client.subscribe(self.data_topic_prefix + topic)

    def add_global_handler(self, topic, handler):
        """
        Add a global handler for a topic.
        The handler will be called with the topic and payload as arguments.
        """

        self.mqtt_global_handlers.append((topic, handler))

        if self.is_connected:
            self.log.info(f"Subscribing to {topic}")
            self.mqtt_client.subscribe(topic)

    def _on_enable(self, payload):
        if payload == "1":
            self.enabled = True
        else:
            self.enabled = False
        self.log.info(f"set enabled to {self.enabled!r}")
        return

    def _update_online_status(self, _):
        self.publish(self.willtopic, "1", retain=True, global_=True)

        assert self.LOOPS

        self.log.info("Loop stats:")
        for l in self.LOOPS:
            call_count, average_call_duration, load, is_critical = l.stat_get()
            l.stat_reset()
            self.log.info(
                f" - {l} called {call_count} times, average duration {average_call_duration}s, load={int(load*100)}%"
            )
            if is_critical:
                self.log.warn(
                    f"   {l} takes, on average, longer to execute ({average_call_duration}s) than defined interval ({l.interval.total_seconds()})"
                )

    def _all_handlers(self):
        for topic, handler in self.mqtt_global_handlers:
            yield topic, handler

        for topic, handler in self.mqtt_handlers:
            yield self.data_topic_prefix + topic, handler

    def _on_message(self, client, userdata, msg):
        payload = str(msg.payload.decode("utf-8", errors="replace")).strip()
        self.log.debug(
            f"Received MQTT message on topic {msg.topic} containing {payload}"
        )

        handled = False
        for topic, handler in self._all_handlers():
            if "#" in topic:
                prefix = topic[:-1]
                self.log.debug(f"matching {prefix} against {msg.topic}")
                if msg.topic.startswith(prefix):
                    handler(self, payload, msg.topic[len(prefix) :])
                    handled = True
            else:
                if topic == msg.topic:
                    handler(self, payload)
                    handled = True

        if handled:
            return

        if self.handle_message(msg.topic, payload):
            return

        self.log.error(
            f"Unhandled topic '{msg.topic}', registered handlers for: {', '.join(k for (k, v) in self._all_handlers())}"
        )

    def handle_message(self, topic, payload):
        return False

    def publish(
        self,
        ext: str | ha_sensors.Entity,
        message,
        retain=False,
        qos=QOS_MAX_ONCE,
        only_if_changed: Union[bool, timedelta] = False,
        global_=False,
    ):
        ha_sensor = None
        if isinstance(ext, ha_sensors.Entity):
            ha_sensor = ext
            ext = ext.state_topic_postfix

        topic = (self.data_topic_prefix + ext) if not global_ else ext
        # if ext not in self.ignore_recv_topics:
        #    self.ignore_recv_topics.append(ext)
        if type(message) == type(True):  # type is boolean
            if ha_sensor is None or not isinstance(ha_sensor, ha_sensors.Switch):
                message = self.PAYLOAD_ON if message else self.PAYLOAD_OFF
            else:
                message = ha_sensor.payload_on if message else ha_sensor.payload_off
        elif message is None:
            message = ""
        elif type(message) in [dict, list]:
            self.publish_json(ext, message, retain, qos, only_if_changed)
            return
        else:
            message = self._round_floats(message)

        if only_if_changed is True:
            last_message = self.last_key_values.get(topic, None)
            if last_message == message:
                self.log.debug(f"{topic} not changed, not publishing.")
                return
            else:
                self.last_key_values[topic] = message
        elif isinstance(only_if_changed, timedelta):
            now = datetime.now()
            last_message, last_time = self.last_key_values.get(topic, (None, None))
            if (
                last_message == message
                and last_time
                and last_time + only_if_changed > now
            ):
                self.log.debug(
                    f"{topic} not changed since {only_if_changed.total_seconds()}s, not publishing."
                )
                return
            else:
                self.last_key_values[topic] = (message, now)

        self.log.debug(f"MQTT publish: {topic}: {message}")
        try:
            self.mqtt_client.publish(topic, message, retain=retain, qos=qos)
        except Exception as e:
            self.log.exception(e)

    def publish_json(
        self,
        ext,
        message_json,
        retain=False,
        qos=QOS_MAX_ONCE,
        only_if_changed: Union[bool, timedelta] = False,
        global_=False,
    ):
        self.publish(
            ext,
            json.dumps(self._round_floats(message_json)),
            retain=retain,
            only_if_changed=only_if_changed,
            global_=global_,
        )

    def publish_json_keys(
        self,
        message_dict: Dict,
        ext=None,
        retain=False,
        qos=QOS_MAX_ONCE,
        only_if_changed: Union[bool, timedelta] = False,
        global_=False,
    ):
        for key, value in message_dict.items():
            if ext:
                key = ext + "/" + key
            # print(key, type(value))
            if type(value) is dict:
                self.publish_json_keys(
                    value, key, retain, qos, only_if_changed, global_
                )
            else:
                self.publish(
                    key,
                    value,
                    retain=retain,
                    only_if_changed=only_if_changed,
                    global_=global_,
                )

    def _loop_step(self):
        assert self.LOOPS is not None

        earliest_next_call = datetime.now() + timedelta(seconds=self.MAX_LOOP_INTERVAL)
        for loop in self.LOOPS:
            next_call = loop.run_if_needed(self)
            if next_call:
                earliest_next_call = min(earliest_next_call, next_call)

        sleep(max(0, (earliest_next_call - datetime.now()).total_seconds()))

    def run(self):
        self.mqtt_client.loop_start()
        while not self.stop:
            self._loop_step()
        self.mqtt_client.loop_stop()

    def _round_floats(self, o):
        if isinstance(o, float):
            return round(o, self.JSON_FLOAT_PRECISION)
        if isinstance(o, dict):
            return {k: self._round_floats(v) for k, v in o.items()}
        if isinstance(o, (list, tuple)):
            return [self._round_floats(x) for x in o]
        return o


def run(service):
    parser = argparse.ArgumentParser(
        description=f"{service.SERVICE_NAME} MIQRO service"
    )
    parser.add_argument("--config", "-c", help="config file", default=None)
    parser.add_argument(
        "--install", action="store_true", help="Setup this service as a systemd unit."
    )
    parser.add_argument(
        "--install-as-user",
        "-u",
        help="Install service for specified user (instead of root)",
        default="root",
    )
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument(
        "--mqtt-debug-prefix",
        "-d",
        help="Prefix for all outgoing MQTT messages for debugging purposes",
        default=None,
    )
    cli_args = parser.parse_args()

    class DebugMQTTClient(mqtt.Client):
        def publish(self, topic, *args, **kwargs):
            return super().publish(cli_args.mqtt_debug_prefix + topic, *args, **kwargs)

        def will_set(self, topic, *args, **kwargs):
            return super().will_set(cli_args.mqtt_debug_prefix + topic, *args, **kwargs)

    if not cli_args.install:
        service(
            cli_args.config,
            logging.DEBUG if cli_args.verbose else logging.INFO,
            DebugMQTTClient if cli_args.mqtt_debug_prefix else mqtt.Client,
        ).run()
        return

    systemd_service_name = f"miqro_{service.SERVICE_NAME}"

    executable = sys.executable
    if not executable:
        executable = "/usr/bin/env python3"

    # Check if the service was started with 'python -m module'
    main_spec = getattr(sys.modules.get("__main__"), "__spec__", None)
    if main_spec and main_spec.name and main_spec.name != "__main__":
        exec_start = f"{executable} -m {main_spec.name}"
    else:
        exec_start = f"{executable} {Path(sys.argv[0]).resolve()}"

    systemd_unit_file = f"""
[Unit]
Description={service.SERVICE_NAME} MIQRO microservice
After=network.target

[Service]
Type=simple
Restart=always
RestartSec=20
User={cli_args.install_as_user}
ExecStart={exec_start}

[Install]
WantedBy=multi-user.target
"""

    systemd_path = Path("/etc/systemd/system/", systemd_service_name + ".service")
    systemd_path.write_text(systemd_unit_file)
    systemd_path.chmod(0o644)
    print(
        f"Service successfully installed as {systemd_service_name}.\nYou can now enable the service to start on boot by running:\n sudo systemctl enable {systemd_service_name}\n... and run the service:\n sudo systemctl start {systemd_service_name}"
    )



from dataclasses import dataclass, field
import re


def clean_string(raw: str) -> str:
    """
    MQTT Discovery protocol only allows [a-zA-Z0-9_-]
    """
    result = re.sub(r"[^A-Za-z0-9_-]", "-", raw)
    return result.strip("-")


@dataclass
class Device:
    service: "miqro.Service"
    name: str
    model: str | None = None
    manufacturer: str | None = None
    sw_version: str | None = None
    hw_version: str | None = None
    identifiers: list[str] | str | None = None
    connections: list[tuple] | None = None
    suggested_area: str | None = None
    via_device: "Device | None" = None

    _unique_id: str | None = None

    __entities: list["Entity"] = field(default_factory=list)

    def __post_init__(self):
        self.service.ha_devices.append(self)
        self._unique_id = f"{self.service.SERVICE_NAME}_{clean_string(self.name)}"
        if self.identifiers is None:
            self.identifiers = [self._unique_id]
        else:
            if isinstance(self.identifiers, str):
                self.identifiers = [self.identifiers]
            if self._unique_id not in self.identifiers:
                self.identifiers.append(self._unique_id)

    @property
    def entities(self):
        return self.__entities

    def add_entity(self, entity):
        self.__entities.append(entity)

    def publish_discovery(self, prefix):
        device_payload = {
            k: v
            for k, v in self.__dict__.items()
            if v is not None and not k.startswith("_")
        }

        del device_payload["service"]

        if self.via_device is not None:
            device_payload["via_device"] = self.via_device._unique_id

        payload = {
            "device": device_payload,
            "origin": {"name": f"MIQRO service {self.service.SERVICE_NAME}"},
            "availability": [
                {
                    "topic": self.service.willtopic,
                    "payload_available": "1",
                    "payload_not_available": "0",
                }
            ],
            "components": {},
        }

        for entity in self.__entities:
            payload["components"][entity.unique_id] = entity.get_discover_payload()

        topic = f"{prefix}/device/{self._unique_id}/config"
        self.service.publish_json(topic, payload, qos=1, retain=True, global_=True)


@dataclass
class EntityWithoutStateTopic:
    device: Device | None
    name: str
    display_name: str | None = None
    device_class: str | None = None
    enabled_by_default: bool | None = None
    entity_category: str | None = None
    expire_after: int | None = None
    force_update: bool | None = None
    icon: str | None = None
    default_entity_id: str | None = None
    qos: int | None = None
    unique_id: str | None = None
    service: "miqro.Service | None" = None

    json_attributes_topic_postfix: str | None = None
    json_attributes_template: str | None = None

    def __post_init__(self):
        # device OR service may be used, but not both
        if self.device is None and self.service is None:
            raise ValueError("Either device or service must be set for an entity")
        if self.device is not None and self.service is not None:
            raise ValueError("Only one of device or service may be set for an entity")

        if self.default_entity_id is None:
            self.default_entity_id = f"{self._component}.{clean_string(self.name)}"
        if self.unique_id is None:
            if self.device is not None:
                self.unique_id = f"{self.device._unique_id}__{self.default_entity_id.replace('.', '_')}"
            else:
                self.unique_id = f"{clean_string(self.service.SERVICE_NAME)}__{self.default_entity_id.replace('.', '_')}"

        # iterate through all attributes and find any callback functions,
        # x_topic_prefix uses the corresponding x_callback function
        for attr_name in dir(self):
            if attr_name.endswith("_topic_postfix"):
                callback_attr_name = attr_name[:-14] + "_callback"
                topic_postfix = getattr(self, attr_name)
                callback = getattr(self, callback_attr_name, None)
                if callback is not None:
                    self.device.service.add_handler(topic_postfix, callback)

        if self.device is not None:
            self.device.add_entity(self)
        else:
            self.service.ha_entities.append(self)

    def get_discover_payload(self):
        payload = {
            k: v
            for k, v in self.__dict__.items()
            if v is not None
            and not k.startswith("_")
            and k != "device"
            and k != "service"
        }
        payload["platform"] = self._component

        # delete x_callback entries from payload
        for attr_name in dir(self):
            if attr_name.endswith("_callback"):
                if attr_name in payload:
                    del payload[attr_name]

        service = self.device.service if self.device is not None else self.service
        for key, value in list(payload.items()):
            if key.endswith("_postfix"):
                full_key = key[:-8]
                payload[full_key] = f"{service.data_topic_prefix}{value}"
                del payload[key]

        return payload

    def publish_discovery(self, prefix):
        # only if device is none
        if self.device is not None:
            raise ValueError(
                "publish_discovery can only be called for entities without device"
            )

        payload = self.get_discover_payload()
        topic = f"{prefix}/{self._component}/{self.unique_id}/config"
        self.service.publish_json(topic, payload, qos=1, retain=True, global_=True)


@dataclass
class Entity(EntityWithoutStateTopic):
    state_topic_postfix: str = ""
    value_template: str | None = None

    def __post_init__(self):
        super().__post_init__()
        if self.state_topic_postfix == "":
            raise ValueError("state_topic_postfix must be set")
        if self.default_entity_id is None:
            self.default_entity_id = (
                f"{self._component}.{self.state_topic_postfix.replace('/', '_')}"
            )

        super().__post_init__()


@dataclass
class Event(Entity):
    _component: str = "event"
    event_types: list | None = None

    def __post_init__(self):
        super().__post_init__()

        if self.event_type is None:
            self.event_type = self.state_topic_postfix.replace("/", "_")


@dataclass
class BinarySensor(Entity):
    _component: str = "binary_sensor"
    off_delay: int | None = None
    payload_off: str | None = None
    payload_on: str | None = None

    def __post_init__(self):
        super().__post_init__()

        service = self.device.service if self.device is not None else self.service

        if self.payload_on is None:
            self.payload_on = str(service.PAYLOAD_ON)
        if self.payload_off is None:
            self.payload_off = str(service.PAYLOAD_OFF)


@dataclass
class Sensor(Entity):
    _component: str = "sensor"
    unit_of_measurement: str | None = None
    state_class: str | None = None
    last_reset_value_template: str | None = None
    suggested_display_precision: int | None = None
    options: list | None = None

    def __post_init__(self):
        super().__post_init__()

        if self.options is not None:
            self.device_class = "enum"


@dataclass
class Switch(Entity):
    _component: str = "switch"
    optimistic: bool | None = None
    command_topic_postfix: str = ""
    command_callback: "callable | None" = None
    payload_off: str = "off"
    payload_on: str = "on"

    def __post_init__(self):
        super().__post_init__()

        if self.payload_on is None:
            self.payload_on = self.device.service.PAYLOAD_ON
        if self.payload_off is None:
            self.payload_off = self.device.service.PAYLOAD_OFF


@dataclass
class Text(Entity):
    _component: str = "text"
    optimistic: bool | None = None
    command_topic_postfix: str = ""
    command_callback: "callable | None" = None
    max: int = 255
    min: int = 0
    mode: str | None = "text"
    pattern: str | None = None
    retain: bool | None = None


@dataclass
class Number(Entity):
    _component: str = "number"
    optimistic: bool | None = None
    command_topic_postfix: str = ""
    command_callback: "callable | None" = None
    max: float | int = 100
    min: float | int = 1
    mode: str | None = None
    payload_reset: str | None = None
    retain: bool | None = None
    step: float | None = None
    unit_of_measurement: str | None = None


@dataclass
class Select(Entity):
    _component: str = "select"
    optimistic: bool | None = None
    command_topic_postfix: str = ""
    command_callback: "callable | None" = None
    retain: bool | None = None
    options: list = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()

        if not self.options:
            raise ValueError("options must be set for select entity")


@dataclass
class Button(EntityWithoutStateTopic):
    _component: str = "button"
    optimistic: bool | None = None
    command_topic_postfix: str = ""
    command_callback: "callable | None" = None
    payload_press: str = "PRESS"
    retain: bool | None = None


@dataclass
class ClimateController(EntityWithoutStateTopic):
    _component = "climate"
    optimistic: bool | None = None
    current_humidity_template: str | None = None
    current_humidity_topic_postfix: str | None = None

    current_temperature_template: str | None = None
    current_temperature_topic_postfix: str | None = None

    fan_mode_command_template: str | None = None
    fan_mode_command_topic_postfix: str | None = None
    fan_mode_command_callback: "callable | None" = None
    fan_mode_state_topic_postfix: str | None = None
    fan_mode_state_template: str | None = None
    fan_modes: list | None = None

    initial: float | None = None  # initial target temperature

    max_humidity: float | None = None
    max_temp: float | None = None
    min_humidity: float | None = None
    min_temp: float | None = None

    mode_command_template: str | None = None
    mode_command_topic_postfix: str | None = None
    mode_state_template: str | None = None
    mode_state_topic_postfix: str | None = None
    modes: list | None = None

    payload_on: str | None = None
    payload_off: str | None = None
    power_command_callback: "callable | None" = None
    power_command_topic_postfix: str | None = None
    power_command_template: str | None = None

    precision: str | None = None

    preset_mode_command_template: str | None = None
    preset_mode_command_topic_postfix: str | None = None
    preset_mode_command_callback: "callable | None" = None
    preset_mode_state_template: str | None = None
    preset_mode_state_topic_postfix: str | None = None
    preset_modes: list | None = None

    swing_horizontal_mode_command_template: str | None = None
    swing_horizontal_mode_command_topic_postfix: str | None = None
    swing_horizontal_mode_command_callback: "callable | None" = None
    swing_horizontal_mode_state_template: str | None = None
    swing_horizontal_mode_state_topic_postfix: str | None = None
    swing_horizontal_modes: list | None = None

    swing_mode_command_template: str | None = None
    swing_mode_command_topic_postfix: str | None = None
    swing_mode_command_callback: "callable | None" = None
    swing_mode_state_template: str | None = None
    swing_mode_state_topic_postfix: str | None = None
    swing_modes: list | None = None

    target_humidity_command_template: str | None = None
    target_humidity_command_topic_postfix: str | None = None
    target_humidity_command_callback: "callable | None" = None
    target_humidity_state_template: str | None = None
    target_humidity_state_topic_postfix: str | None = None

    temperature_command_topic_postfix: str | None = None
    temperature_command_template: str | None = None
    temperature_command_callback: "callable | None" = None
    temperature_state_template: str | None = None
    temperature_state_topic_postfix: str | None = None

    temperature_high_command_topic_postfix: str | None = None
    temperature_high_command_template: str | None = None
    temperature_high_command_callback: "callable | None" = None
    temperature_high_state_template: str | None = None
    temperature_high_state_topic_postfix: str | None = None

    temperature_low_command_topic_postfix: str | None = None
    temperature_low_command_template: str | None = None
    temperature_low_command_callback: "callable | None" = None
    temperature_low_state_template: str | None = None
    temperature_low_state_topic_postfix: str | None = None

    temperature_unit: str | None = None
    temp_step: float | None = None

    def __post_init__(self):
        super().__post_init__()

        # precision must be one of '0.1', '0.5', '1.0'
        if self.precision is not None:
            if self.precision not in ["0.1", "0.5", "1.0"]:
                raise ValueError("precision must be one of '0.1', '0.5', '1.0'")

        # temperature unit must be C or F
        if self.temperature_unit is not None:
            if self.temperature_unit not in ["C", "F"]:
                raise ValueError("temperature_unit must be 'C' or 'F'")


@dataclass
class DeviceTracker(EntityWithoutStateTopic):
    _component = "device_tracker"
    json_attributes_topic_postfix: str | None = None
    json_attributes_template: str | None = None
